/*******************************************************************************
 * Copyright (c) 2012, 2015 Google, Inc and others.
 *
 * This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License 2.0
 * which accompanies this distribution, and is available at
 * https://www.eclipse.org/legal/epl-2.0/
 *
 * SPDX-License-Identifier: EPL-2.0
 *
 * Contributors:
 * 	   Sergey Prigogin (Google) - initial API and implementation
 *******************************************************************************/
package org.eclipse.cdt.core.parser.tests.ast2;

import java.io.IOException;
import java.util.Optional;

import org.eclipse.cdt.core.dom.ast.IASTName;
import org.eclipse.cdt.core.parser.ParserLanguage;
import org.eclipse.cdt.internal.core.dom.parser.c.CVariableReadWriteFlags;
import org.eclipse.cdt.internal.core.dom.parser.cpp.semantics.CPPVariableReadWriteFlags;
import org.eclipse.cdt.internal.core.parser.ParserException;
import org.eclipse.cdt.internal.core.pdom.dom.PDOMName;

import junit.framework.TestSuite;

/**
 * Unit tests for CPPVariableReadWriteFlags and CVariableReadWriteFlags classes.
 */
public class VariableReadWriteFlagsTest extends AST2TestBase {
	private static final int READ = PDOMName.READ_ACCESS;
	private static final int WRITE = PDOMName.WRITE_ACCESS;

	protected class AssertionHelper extends AST2AssertionHelper {
		AssertionHelper(String contents, boolean isCPP) throws ParserException {
			super(contents, isCPP);
		}

		AssertionHelper(String contents, ParserLanguage lang, ScannerKind scannerKind) throws ParserException {
			super(contents, lang, scannerKind);
		}

		void assertReadWriteFlags(String context, String name, Optional<Integer> expectedFlags) throws Exception {
			IASTName variable = findName(context, name);
			assertNotNull(variable);
			assertEquals(flagsToString(expectedFlags), flagsToString(getReadWriteFlags(variable)));
		}

		void assertReadWriteFlags(String context, String name, int expectedFlags) throws Exception {
			assertReadWriteFlags(context, name, Optional.of(expectedFlags));
		}

		void assertReadWriteFlags(String context, String name) throws Exception {
			assertReadWriteFlags(context, name, Optional.empty());
		}

		Optional<Integer> getReadWriteFlags(IASTName variable) {
			return isCPP ? CPPVariableReadWriteFlags.getReadWriteFlags(variable)
					: CVariableReadWriteFlags.getReadWriteFlags(variable);
		}

		private String flagsToString(Optional<Integer> flags) {
			StringBuilder buf = new StringBuilder();
			if (!flags.isPresent()) {
				buf.append("UNK");
			}
			if (flags.isPresent() && (flags.get() & READ) != 0) {
				buf.append("READ");
			}
			if (flags.isPresent() && (flags.get() & WRITE) != 0) {
				if (buf.length() != 0)
					buf.append(" | ");
				buf.append("WRITE");
			}
			if (buf.length() == 0)
				buf.append("0");
			return buf.toString();
		}
	}

	public VariableReadWriteFlagsTest() {
	}

	public VariableReadWriteFlagsTest(String name) {
		super(name);
	}

	public static TestSuite suite() {
		return suite(VariableReadWriteFlagsTest.class);
	}

	protected AssertionHelper getCAssertionHelper() throws ParserException, IOException {
		String code = getAboveComment();
		return new AssertionHelper(code, false);
	}

	protected AssertionHelper getCPPAssertionHelper() throws ParserException, IOException {
		String code = getAboveComment();
		return new AssertionHelper(code, true);
	}

	protected AssertionHelper getCPP20AssertionHelper() throws ParserException, IOException {
		String code = getAboveComment();
		return new AssertionHelper(code, ParserLanguage.CPP, ScannerKind.STDCPP20);
	}

	//	int test(int a) {
	//	  a = 2;
	//	  a *= 3;
	//	  return a + 1;
	//	}
	public void testSimpleAccess() throws Exception {
		AssertionHelper a = getCPPAssertionHelper();
		a.assertReadWriteFlags("a = 2", "a", WRITE);
		a.assertReadWriteFlags("a *= 3", "a", READ | WRITE);
		a.assertReadWriteFlags("a + 1", "a", READ);
	}

	//	auto test(int a) {
	//	  return a <=> 1;
	//	}
	public void testThreeWayComparisonAccess() throws Exception {
		AssertionHelper a = getCPP20AssertionHelper();
		a.assertReadWriteFlags("a <=> 1", "a", READ);
	}

	//	class C {
	//	public:
	//	  C(int);
	//	};
	//
	//	class D {
	//	public:
	//	  D();
	//	};
	//
	//	int a;
	//	int b = 1;
	//	C c;
	//	D d;
	//	C e(1);
	//	template<typename T> void foo(T p) {
	//	  T f;
	//	}
	public void testVariableDeclaration() throws Exception {
		AssertionHelper a = getCPPAssertionHelper();
		a.assertReadWriteFlags("int a", "a", 0);
		a.assertReadWriteFlags("int b = 1", "b", WRITE);
		a.assertReadWriteFlags("C c", "c", 0);
		a.assertReadWriteFlags("D d", "d", WRITE);
		a.assertReadWriteFlags("C e(1)", "e", WRITE);
		a.assertReadWriteFlags("T f", "f", WRITE);
	}

	//	struct A { int x; };
	//
	//	void test(A a, A* ap) {
	//	  a.x = 1;
	//	  (&a)->x = 1;
	//	  ap->x = 1;
	//	};
	public void testFieldAccess() throws Exception {
		AssertionHelper a = getCPPAssertionHelper();
		a.assertReadWriteFlags("a.x", "a", WRITE);
		a.assertReadWriteFlags("a.x", "x", WRITE);
		a.assertReadWriteFlags("(&a)->x", "a", WRITE);
		a.assertReadWriteFlags("(&a)->x", "x", WRITE);
		a.assertReadWriteFlags("ap->x", "ap", READ);
		a.assertReadWriteFlags("ap->x", "x", WRITE);
	}

	//	void f(int* x, int& y);
	//	void g(const int* x, const int& y, int z);
	//
	//	void test(int a, int b, int c) {
	//	  f(&a, b);
	//	  g(&a, b, c);
	//	};
	public void testFunctionCall() throws Exception {
		AssertionHelper a = getCPPAssertionHelper();
		a.assertReadWriteFlags("f(&a, b)", "a");
		a.assertReadWriteFlags("f(&a, b)", "b");
		a.assertReadWriteFlags("f(&a, b)", "f", READ);
		a.assertReadWriteFlags("g(&a, b, c)", "a", READ);
		a.assertReadWriteFlags("g(&a, b, c)", "b", READ);
		a.assertReadWriteFlags("g(&a, b, c)", "c", READ);
	}

	//	struct A {
	//	  A(int* x, int& y);
	//	  A(const int* x, const int& y, int z);
	//	};
	//
	//	void test(int a, int b, int c) {
	//	  A u = A(&a, b);
	//	  A* v = new A(&a, b);
	//	  A w(&a, b);
	//	  A x = A(&a, b, c);
	//	  A* y = new A(&a, b, c);
	//	  A z(&a, b, c);
	//	};
	public void testConstructorCall_393068() throws Exception {
		AssertionHelper a = getCPPAssertionHelper();
		a.assertReadWriteFlags("= A(&a, b)", "a");
		a.assertReadWriteFlags("= A(&a, b)", "b");
		a.assertReadWriteFlags("new A(&a, b)", "a");
		a.assertReadWriteFlags("new A(&a, b)", "b");
		a.assertReadWriteFlags("w(&a, b)", "a");
		a.assertReadWriteFlags("w(&a, b)", "b");
		a.assertReadWriteFlags("w(&a, b)", "w", WRITE);
		a.assertReadWriteFlags("= A(&a, b, c)", "a", READ);
		a.assertReadWriteFlags("= A(&a, b, c)", "b", READ);
		a.assertReadWriteFlags("= A(&a, b, c)", "c", READ);
		a.assertReadWriteFlags("new A(&a, b, c)", "a", READ);
		a.assertReadWriteFlags("new A(&a, b, c)", "b", READ);
		a.assertReadWriteFlags("new A(&a, b, c)", "c", READ);
		a.assertReadWriteFlags("z(&a, b, c)", "a", READ);
		a.assertReadWriteFlags("z(&a, b, c)", "b", READ);
		a.assertReadWriteFlags("z(&a, b, c)", "c", READ);
	}

	//	struct A {
	//	  void m();
	//	  void mc() const;
	//	};
	//
	//	void test(A a, A* ap) {
	//	  a.m();
	//	  a.mc();
	//	  (&a)->m();
	//	  (&a)->mc();
	//	  ap->m();
	//	  (*ap).m();
	//	};
	public void testMethodCall() throws Exception {
		AssertionHelper a = getCPPAssertionHelper();
		a.assertReadWriteFlags("a.m()", "a", READ | WRITE);
		a.assertReadWriteFlags("a.m()", "m", READ);
		a.assertReadWriteFlags("a.mc()", "a", READ);
		a.assertReadWriteFlags("(&a)->m()", "a", READ | WRITE);
		a.assertReadWriteFlags("(&a)->m()", "m", READ);
		a.assertReadWriteFlags("ap->m()", "ap", READ);
		a.assertReadWriteFlags("(*ap).m()", "ap", READ);
	}

	//	void variadic(...);
	//	void test() {
	//		int waldo;
	//		variadic(waldo);
	//		variadic(&waldo);
	//	}
	public void testVariadicFunctionCall_452416() throws Exception {
		AssertionHelper a = getCPPAssertionHelper();
		a.assertReadWriteFlags("variadic(waldo)", "waldo", READ);
		a.assertReadWriteFlags("variadic(&waldo)", "waldo");
	}

	//	int arr[5];
	//	int test() {
	//		arr[0];
	//		arr[0] = 1;
	//		int a;
	//		a = arr[0];
	//		return arr[0];
	//	}
	public void testArraySubscript() throws Exception {
		AssertionHelper a = getCPPAssertionHelper();
		a.assertReadWriteFlags("arr[0];", "arr", READ);
		a.assertReadWriteFlags("arr[0] = 1", "arr", READ | WRITE);
		a.assertReadWriteFlags("a = arr[0];", "arr", READ);
		a.assertReadWriteFlags("return arr[0];", "arr", READ);
	}

	//	class Test {
	//  public:
	//     int v;
	//     void test() const {
	//         decltype(v) o = 14;
	//     }
	//  };
	public void testDeclType() throws Exception {
		AssertionHelper a = getCPPAssertionHelper();
		a.assertReadWriteFlags("decltype(v) o = 14;", "v", READ);
	}

	//template <typename S, typename T>
	//void g(S, T);
	//
	//struct A {
	//	int field;
	//
	//	template <typename S>
	//	void method3(S s) const {
	//  	  g(s, field);
	//	}
	//};
	public void testDependentType() throws Exception {
		AssertionHelper a = getCPPAssertionHelper();
		a.assertReadWriteFlags("g(s, field);", "field");
	}
}
